/*******************************************************************************
 * Copyright (C) 2024
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
#include "ProcessingElementCS.h"
#include "traffic/PacketCS.h"
#include "utils/noc_logger.h"

void ProcessingElementCS::thread(){
    while(true) {
        int timeStamp = static_cast<int>(sc_time_stamp().value()/1000);
        std::vector<DataDestinationCS> removeList;
        // EXPLAIN: check if time of the task has already been passed. If it has passed remove it.
        for (auto const& tw : destWait) {
            DataDestinationCS dest = tw.first;
            TaskCS task = destToTask.at(dest);
            if (taskTerminationTime.count(task) && taskTerminationTime.at(task)<timeStamp) {
                removeList.push_back(dest);
            }
        }

        // EXPLAIN: removing task
        for (auto& dest : removeList) {
            log_info("Removing task " + std::to_string(destToTask.at(dest).id));
            TaskCS task = destToTask.at(dest);
            destToTask.erase(dest);
            taskToDest.erase(task);
            taskRepeatLeft.erase(task);
            taskStartTime.erase(task);
            taskTerminationTime.erase(task);
            countLeft.erase(dest);
            destWait.erase(dest);
        }

        // EXPLAIN: loop through all destination that are waiting
        for (auto const& pair : destWait) {
            DataDestinationCS dest = pair.first;
            // EXPLAIN: check if the current time is smaller or equal to the time of the destination if yes generate packet
            if (pair.second<=timeStamp) {
                std::string curr_task_id = std::to_string(destToTask.at(dest).id);
                TaskCS task = globalResources.tasks.at(dest.destinationTask);
                Node dstNode = globalResources.nodes.at(task.nodeID);
                PacketCS* p = packetFactoryCS.createPacket(this->node, 
                    dstNode, globalResources.flitsPerPacket, 
                    sc_time_stamp().to_double(), dest.dataType, dest.msg);
                double time = sc_time_stamp().to_double();

                // EXPLAIN: create packet and send it
                if ((float) globalResources.synthetic_start_measurement_time<=(time/1000))
                    globalReport.undeliveredPackages++;

                packetPortContainer->portValidOut = true;
                packetPortContainer->portDataOut = p;

                // EXPLAIN: decrement number of packets left to be sent
                // Explain: countLeft is for sending, neededAmount is for receiving
                countLeft.at(dest)--;
                log_send_rec("Task " + curr_task_id + " sent a packet of type " +
                        std::to_string(dest.dataType) + " (msg: " +
                        std::to_string(dest.msg.config_link) + "-" +
                        std::to_string(dest.msg.dest_link) + ")" +
                        " destined for task " + 
                        std::to_string(dest.destinationTask) + 
                        " (repeat count: " + 
                        std::to_string(taskRepeatLeft.at(destToTask.at(dest))) + 
                        ")");

                // EXPLAIN: if there are no more packet to this destination remove destination from list of destinations
                if (!countLeft.at(dest)) {
                    countLeft.erase(dest);
                    destWait.erase(dest);
                    log_info("Task " + curr_task_id + 
                        " has sent all packets to destination " +
                        std::to_string(task.id));

                    TaskCS task = destToTask.at(dest);
                    destToTask.erase(dest);
                    taskToDest.at(task).erase(dest);

                    // EXPLAIN: There are no more destinations to the task, all packets of the task has been send
                    if (taskToDest.at(task).empty()) {
                        taskToDest.erase(task);
                        execute(task);
                    }
                }
                // EXPLAIN: is there are more packets to the destination schedule the next packet
                else {
                    destWait.at(dest) =
                            globalResources.getRandomIntBetween(dest.minInterval, dest.maxInterval)+timeStamp;
                }                
                break;
            }
        }

        wait(SC_ZERO_TIME);
        packetPortContainer->portValidOut = false;

        int nextCall = -1;
        for (auto const& dw : destWait) {
            if (nextCall>dw.second || nextCall==-1) {
                if (globalResources.benchmark=="synthetic") {
                    TaskCS task = this->destToTask.at(dw.first);
                    int minInterval = dw.first.minInterval;

                    if (timeStamp<task.minStart) {
                        nextCall = task.minStart+globalResources.getRandomIntBetween(0, minInterval-1);
                    }
                    else {
                        int numIntervalsPassed = (timeStamp-task.minStart)/minInterval;
                        int intervalBeginning = task.minStart+(numIntervalsPassed*minInterval);
                        nextCall = intervalBeginning+minInterval+globalResources.getRandomIntBetween(0, minInterval-1);
                    }
                }
                else { // if not synthetic, then execute the original behavior
                    nextCall = dw.second;
                }
            }
        }

        if (nextCall==0) { // limit packet rate
            nextCall = 1;
        }
        if (nextCall!=-1) {
            event.notify(nextCall-timeStamp, SC_NS);
        }
        wait(event);
    }
}


void ProcessingElementCS::execute(TaskCS& task)
{
    // EXPLAIN: if there are no repetitions of the task left, create a new number of repetitions (probably used only in the first call for init)
    if (!taskRepeatLeft.count(task)) {
        taskRepeatLeft[task] = globalResources.getRandomIntBetween(task.minRepeat, task.maxRepeat);
    }
    else {
        // EXPLAIN: decrement the number of repetitions left by 1 and if they reach 0 after that remove the task from list (task is done).
        if (taskRepeatLeft.at(task)>0) {
            taskRepeatLeft.at(task)--;
            log_info("Task " + std::to_string(task.id) +
                " has completed one repitition. Repetitions left: " + 
                std::to_string(taskRepeatLeft.at(task)));
        }

        if (!taskRepeatLeft.at(task)) {
            taskRepeatLeft.erase(task);
            log_info("Task " + std::to_string(task.id) + " has completed all repetitions");
            return;
        }
    }

    // EXPLAIN: if there is no start time init one
    if (!taskStartTime.count(task)) {
        taskStartTime[task] = globalResources.getRandomIntBetween(task.minStart, task.maxStart);
    }

    // EXPLAIN: if there is no termination time init one
    if (!taskTerminationTime.count(task) && task.minDuration!=-1) {
        taskTerminationTime[task] =
                taskStartTime[task]+globalResources.getRandomIntBetween(task.minDuration, task.maxDuration);
    }

    // EXPLAIN: if there are no requirements left start sending
    if (task.requirements.empty()) {
        startSending(task);
    }
    else {
        // EXPLAIN: go through all requirements and add them to lists
        for (DataRequirement& r : task.requirements) {
            log_info("Requirement of id " + std::to_string(r.id) +
                    " of type " + std::to_string(r.dataType) +
                    " created for task " + std::to_string(task.id));
            neededFor[r.dataType].insert(task);
            int amount_req = globalResources.getRandomIntBetween(r.minCount, r.maxCount);
            neededAmount[std::make_pair(task, r.dataType)] = amount_req;
            //neededAmountHold[std::make_pair(task, r.dataType)] = amount_req;
            needs[task].insert(r.dataType);
        }
        // maybe all packets have arrived so we check the requirements
        checkNeed();
    }
}



void ProcessingElementCS::receive()
{
    LOG(globalReport.verbose_pe_function_calls,
            "PE" << this->id << "(Node" << node.id << ")\t- receive_data_process()");

    // EXPLAIN: checks for positive edge of the validIn signal. This is the trigger
    if (packetPortContainer->portValidIn.posedge()) {

        // EXPLAIN: read packet from port
        Packet* received_packet = packetPortContainer->portDataIn.read();

        if (received_packet) {
            dataTypeID_t type = received_packet->dataType;

            // EXPLAIN: check if receivedData already has a counter for the packet type of the received packet. If yes increment if, if no create one
            if (receivedData.count(type)) {
                ++receivedData.at(type);
            }
            else {
                receivedData[type] = 1;
            }
            log_send_rec("Packet from type " + std::to_string(type) + 
                    " arrived at PE " + std::string(this->name()) + " from node " + 
                    std::to_string(received_packet->src.id) + 
                    " - current received packet counter: " + std::to_string(receivedData[type]));

            // EXPLAIN: check if now all required data has arrived and if yes start sending data
            checkNeed();

            // EXPLAIN: delete packet
            packetFactoryCS.deletePacket(received_packet);
        }
    }
}

// EXPLAIN: select possible destinations and set all timings and packet numbers for these destinations they are saved in destWait
void ProcessingElementCS::startSending(TaskCS& task)
{
    // EXPLAIN: There are multiple possibilities where to send the data. Select one based on probability
    float rn = globalResources.getRandomFloatBetween(0, 1);
    int numOfPoss = task.possibilities.size();
    int numOfDests = 0;
    if(numOfPoss) {
        numOfDests = task.possibilities.at(0).dataDestinations.size();
    }
    // for tasks with no destinations, only requirements
    if (!numOfDests) {
        taskRepeatLeft.at(task)--;
        if (taskRepeatLeft.at(task)){
            for (DataRequirement& r : task.requirements) {
                log_info("Requirement of id " + std::to_string(r.id) +
                        " of type " + std::to_string(r.dataType) +
                        " created for task " + std::to_string(task.id));
                neededFor[r.dataType].insert(task);
                int amount_req = globalResources.getRandomIntBetween(r.minCount, r.maxCount);
                neededAmount[std::make_pair(task, r.dataType)] = amount_req;
                //neededAmountHold[std::make_pair(task, r.dataType)] = amount_req;
                needs[task].insert(r.dataType);
            }
            checkNeed();
        }
    }

    for (unsigned int i = 0; i<numOfPoss; ++i) {
        if (task.possibilities.at(i).probability>rn) {
            // EXPLAIN: A possibility is chosen. Set the destinations of this possibility as the dest of the task
            std::vector<DataDestinationCS> destVec = task.possibilities.at(i).dataDestinations;
            for (DataDestinationCS& dest : destVec) {
                destToTask[dest] = task;
                taskToDest[task].insert(dest);

                // EXPLAIN: get random packet count per dest
                countLeft[dest] = globalResources.getRandomIntBetween(dest.minCount, dest.maxCount);

                // EXPLAIN: random delay time to start sending data to dest
                int delayTime = static_cast<int>((sc_time_stamp().value()/1000)
                                +globalResources.getRandomIntBetween(dest.minDelay, dest.maxDelay));

                if (taskStartTime.count(task) && taskStartTime.at(task)>delayTime) {
                    destWait[dest] = taskStartTime.at(task);
                }
                else {
                    destWait[dest] = delayTime;
                }
                log_info("Task " + std::to_string(task.id) + " with destination of id " +
                    std::to_string(dest.id) + " and task id " +
                    std::to_string(dest.destinationTask) + " of type " + 
                    std::to_string(dest.dataType) + " created");
                // EXPLAIN: schedule event in systemC
                event.notify(SC_ZERO_TIME);
            }
            break;
        }
        else {
            rn -= task.possibilities.at(i).probability;
        }
    }
}

void ProcessingElementCS::checkNeed()
{
    // EXPLAIN: iterate through received data
    for (auto const& data : receivedData) {
        dataTypeID_t type = data.first;
        std::vector<std::pair<TaskCS, dataTypeID_t>> removeList;

        // EXPLAIN: check if there are task that need this data
        if (neededFor.count(type)) {

            // EXPLAIN: iterate over tasks needing the data 
            for (const TaskCS& t : neededFor.at(type)) {
                std::pair<TaskCS, dataTypeID_t> pair = std::make_pair(t, type);
                /* EXPLAIN: commented because received data should always 
                count to one, the purpose of this line is unclear, but it caused
                an error where the requirement in a task wouldn't be read */
                // uncommented because data that had arrived wasn't diminishing from neededAmount counter
                neededAmount.at(pair) -= receivedData.at(type);
                log_send_rec("Received data of type " + std::to_string(type)+
                    " for task " + std::to_string(t.id) + ": " + 
                    std::to_string(receivedData.at(type)));
                //neededAmount.at(pair)--;
                /* This line was commented out because if a task requires several packets from several data types,
                 it says that the task is finished receiving the required packets while in fact, it still needs some packets.
                 receivedData.at(type) = 0;
                 */
                 receivedData.at(type) = 0;
                log_send_rec("Needed amount (of packets) of type " + std::to_string(type)+
                    " for task " + std::to_string(t.id) + " updated - current value: " + 
                    std::to_string(neededAmount.at(pair)));

                // EXPLAIN: check if the needed amount (datatype for specific task) is reached. If yes remove the pair
                if (neededAmount.at(pair)<=0) {
                    removeList.push_back(pair);
                    log_info("Task "+std::to_string(t.id)+" has requirement of type "+
                            std::to_string(type)+" fullfilled");
                    // This line is also commented out for the same reason mentioned above.
                    // uncommented because it failed with mobilenet and now it works
                    // Explain: needed amount keeps being reduced, the negative needed amount is the new received data
                    // but it doesn't go to the for and if so it never arrives to the --
                    // so neededAmountHold was created
                    receivedData.at(type) = -neededAmount.at(pair);
                    //receivedData.at(type) -= neededAmountHold.at(pair); // this only works if recevied data has arrived to neededAmountHold before repetition
                }
            }
        }
        // EXPLAIN: remove data from lists
        for (auto& p : removeList) {
            neededFor.erase(p.second);
            neededAmount.erase(p);
            needs.at(p.first).erase(p.second);

            // EXPLAIN: check if all data requirements of the task are satisfied. If yes start sending
            if (needs.at(p.first).empty()) {
                log_info("Task "+std::to_string(p.first.id)+" has all requirements fullfilled");
                startSending(p.first);
            }
        }
    }
}


/******************* LOG FUNCTIONS ********************/
void ProcessingElementCS::log_info(std::string msg){
    SC_REPORT_INFO(PE_LOG, ("PE->" + std::string(this->name()) + ": " + msg).c_str());
}

void ProcessingElementCS::log_warn(std::string msg){
    SC_REPORT_WARNING(PE_LOG, ("PE->" + std::string(this->name()) + ": " +  msg).c_str());
}

void ProcessingElementCS::log_error(std::string msg){
    SC_REPORT_ERROR(PE_LOG, ("PE->" + std::string(this->name()) + ": " +  msg).c_str());
}

void ProcessingElementCS::log_send_rec(std::string msg){
    if (DEBUG) {
        SC_REPORT_INFO(PE_PG_LOG, ("PE->" + std::string(this->name()) + ": " + msg).c_str());
    }
}

